/* SPDX-License-Identifier: BSD-3-Clause
 * Copyright(c) 2019-2021 Broadcom
 * All rights reserved.
 */

#ifndef _TF_RESOURCES_H_
#define _TF_RESOURCES_H_

#define TF_NUM_TBL_SCOPE           16      /* < Number of TBL scopes */

#endif /* _TF_RESOURCES_H_ */
