/* SPDX-License-Identifier: BSD-3-Clause
 * Copyright(C) 2021 Marvell.
 */

#ifndef _CNXK_BPHY_CGX_H_
#define _CNXK_BPHY_CGX_H_

int cnxk_bphy_cgx_dev_selftest(uint16_t dev_id);

#endif /* _CNXK_BPHY_CGX_H_ */
