/* SPDX-License-Identifier: BSD-3-Clause
 * Copyright(c) 2001-2021 Intel Corporation
 */

#ifndef _IAVF_DEVIDS_H_
#define _IAVF_DEVIDS_H_

/* Vendor ID */
#define IAVF_INTEL_VENDOR_ID		0x8086

/* Device IDs for the VF driver */
#define IAVF_DEV_ID_VF			0x154C
#define IAVF_DEV_ID_VF_HV		0x1571
#define IAVF_DEV_ID_ADAPTIVE_VF		0x1889
#define IAVF_DEV_ID_X722_VF		0x37CD
#define IAVF_DEV_ID_X722_A0_VF          0x374D

#endif /* _IAVF_DEVIDS_H_ */
